#include "gdTiff.H"
#include <iostream>
#include <tiffio.h>
#include <cstdlib>
#include <fstream>
#include <cmath>
#include <string>

using namespace std;

#define OUTPUT_ALPHA

float** readChannel(char* fname, int &width, int &height);


TIFF *quick_opentiffr (const char *pic, int *width,
       int *height, unsigned short *samples,
       unsigned short *bitspersample,
       unsigned short *sampleformat,
       uint32 *tilewidth,
       uint32 *tilelength)
{
    TIFFSetWarningHandler(NULL);  // don't need no stinking warnings
    TIFF *tif = TIFFOpen (pic, "r");
    if (! tif) {
	return NULL;
    }
    TIFFGetField (tif, TIFFTAG_IMAGEWIDTH, width);
    TIFFGetField (tif, TIFFTAG_IMAGELENGTH, height);
    *sampleformat = SAMPLEFORMAT_UINT;
    TIFFGetField (tif, TIFFTAG_SAMPLEFORMAT, sampleformat);
    TIFFGetFieldDefaulted (tif, TIFFTAG_BITSPERSAMPLE, bitspersample);
    TIFFGetFieldDefaulted (tif, TIFFTAG_SAMPLESPERPIXEL, samples);
    if (*bitspersample != 8 && *bitspersample != 16 && *bitspersample != 32) {
	return NULL;
    }
    if (tilewidth && tilelength) {
	*tilewidth = *tilelength = 0;
	if (TIFFIsTiled (tif)) {
	    TIFFGetField (tif, TIFFTAG_TILEWIDTH, tilewidth);
	    TIFFGetField (tif, TIFFTAG_TILELENGTH, tilelength);
	}
    }
    return tif;
}



unsigned char *
quick_readtiff(const char *pic, int *width, int *height,
		unsigned short *samples,
		unsigned short *bitspersample,
		unsigned short *sampleformat)
{
    uint32 tilewidth=0, tilelength=0;

    TIFF *tif = quick_opentiffr (pic, width, height, samples, bitspersample,
				sampleformat, &tilewidth, &tilelength);
    if (! tif)
	return NULL;

    int samplesize = (*bitspersample) / 8;
    size_t pixellen = (*samples) * samplesize;
    size_t linelen = (*width) * pixellen;
    size_t buflen = (*height) * linelen;

    unsigned char *raster = new unsigned char [buflen];

	// Scanline image
	for (int y = 0;  y < *height;  ++y)
	TIFFReadScanline (tif, raster + y*linelen, y);

    TIFFClose (tif);

    return raster;
}


double* GdTiff::LoadFloatTiff(char *filename) {
	unsigned short samples, bitspersample, sampleformat;

	data = (double *) quick_readtiff(filename, &_width, &_height, &samples, &bitspersample, &sampleformat);

	_samplesperpixel = (int) samples;
	_bitspersample = (int) bitspersample;
	_sampleformat = sampleformat;
        
	return data;
}


TIFF *
 quick_opentiffw (const char *imagefile, uint32 xres,
                  uint32 yres, short samples, short bitspersample,
                  short sampleformat, unsigned short compression,
                  uint32 tilewidth, uint32 tilelength)
 {
     TIFF *out = TIFFOpen (imagefile, "w");
     if (! out) {
         return NULL;
     }
     TIFFSetField (out, TIFFTAG_IMAGEWIDTH, xres);
     TIFFSetField (out, TIFFTAG_IMAGELENGTH, yres);
     TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
     TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, samples);
     TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, bitspersample);
     TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, sampleformat);
     TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
     TIFFSetField (out, TIFFTAG_COMPRESSION, compression);
     if (tilewidth && tilelength) {
         TIFFSetField (out, TIFFTAG_TILEWIDTH, tilewidth);
         TIFFSetField (out, TIFFTAG_TILELENGTH, tilelength);
     } else {
         TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, 8);
     }
     TIFFSetField (out, TIFFTAG_PHOTOMETRIC, 
                   (samples > 1) ? PHOTOMETRIC_RGB : PHOTOMETRIC_MINISBLACK);
     if (samples == 4) {
         unsigned short sampleinfo = EXTRASAMPLE_ASSOCALPHA;
         TIFFSetField (out, TIFFTAG_EXTRASAMPLES, 1, &sampleinfo);
     }
     return out;
 }



 int
 quick_writetiff (const char *imagefile, const unsigned char *raster,
                  uint32 xres, uint32 yres,
                  short samples, short bitspersample, short sampleformat,
                  unsigned short compression,
                  uint32 tilewidth, uint32 tilelength)
 {
     TIFF *tif = quick_opentiffw (imagefile, xres, yres, samples,
                                  bitspersample, sampleformat, compression,
                                  tilewidth, tilelength);
     if (! tif)
         return -1;
     int samplesize = bitspersample / 8;
     size_t pixellen = samples * samplesize;
     size_t linelen = xres * pixellen;
     if (tilewidth) {
         // Tile file
         unsigned char *tilebuf = (unsigned char *)alloca (TIFFTileSize (tif));
         size_t tilewidlen = pixellen * tilewidth;
         for (int y = 0;  y < yres;  y += tilelength) {
             for (int x = 0;  x < xres;  x += tilewidth) {
                 for (int i = 0;  i < tilelength;  ++i)
                     memcpy (tilebuf+i*tilewidlen,
                             raster+(y+i)*linelen + x*pixellen, tilewidlen);
                 TIFFWriteTile (tif, tilebuf, x, y, 0, 0);
             }
         }
         
     } else {
         // Scanline file
         unsigned char *slbuf = (unsigned char *)alloca (linelen);
         for (int y = 0;  y < yres;  ++y) {
             memcpy (slbuf, raster+y*linelen, linelen);
             TIFFWriteScanline (tif, slbuf, y);
         }
     }
     TIFFClose (tif);
     return 0;
 }


void GdTiff::WriteFloatTiff(char *filename) {

  quick_writetiff(filename, (unsigned char *) data,
                  (uint32) _width, (uint32) _height,
                  (short) _samplesperpixel, (short) _bitspersample, (short) _sampleformat,
                  COMPRESSION_LZW,
                  0, 0);
}

GdTiff::GdTiff(char *filename) {
  int x,y,c;
  if (strstr(filename, "tif")) {
    TIFF* tiff = TIFFOpen(filename, "r");
    TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &_width);
    TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &_height);
    _type=RGB;
    uint32* raster = new uint32[_width*_height];
    TIFFReadRGBAImage(tiff,_width, _height, raster, 1);
    data = new double[_width*_height*4];
    for (y=0; y<_height; y++) {
      for (x=0; x<_width; x++) {
	data[4*x+4*y*_width] = ((double) TIFFGetR(raster[y*_width+x]))/255.0;
	data[4*x+4*y*_width+1] = ((double) TIFFGetG(raster[y*_width+x]))/255.0;
	data[4*x+4*y*_width+2] = ((double) TIFFGetB(raster[y*_width+x]))/255.0;
	data[4*x+4*y*_width+3] = ((double) TIFFGetA(raster[y*_width+x]))/255.0;
	//data[4*x+4*y*_width] = data[4*x+4*y*_width]*data[4*x+4*y*_width];
	//data[4*x+4*y*_width+1] = data[4*x+4*y*_width+1]*data[4*x+4*y*_width+1];
	//data[4*x+4*y*_width+2] = data[4*x+4*y*_width+2]*data[4*x+4*y*_width+2];
      }
    }
    TIFFClose(tiff);
    delete raster;
  } else {
    char fname[80];
    sprintf(fname, "%s.r", filename);
    float** r = readChannel(fname, _width, _height);
    sprintf(fname, "%s.g", filename);
    float** g = readChannel(fname, _width, _height);
    sprintf(fname, "%s.b", filename);
    float** b = readChannel(fname, _width, _height);
    data = new double[_width*_height*4];
    for (y=0; y<_height; y++) 
      for (x=0; x<_width; x++) {
	data[4*x+4*y*_width] = r[_height-y-1][x];
	data[4*x+4*y*_width+1] = g[_height-y-1][x];
	data[4*x+4*y*_width+2] = b[_height-y-1][x];
	data[4*x+4*y*_width+3] = 1.0;
      }
    for (y=0; y<_height; y++) {
      delete[] (r[y]);
      delete[] (g[y]);
      delete[] (b[y]);
    }
    delete [] r;
    delete [] g;
    delete [] b;
  }
}



GdTiff::GdTiff(int width, int height) {
  int x,y;
  _width = width;
  _height = height;
  data = new double[width*height*4];
  _type = UNDEF;
  cerr<<"Warning: creating image with an undefined color space"<<endl;
};



GdTiff::GdTiff(int width, int height, short type) {
  int x,y;
  _width = width;
  _height = height;
  data = new double[width*height*4];
  this->_type = type;
};



GdTiff::~GdTiff() {
  delete [] data;
}

void GdTiff::write_float(char *fname) {
  TIFF *file;
  float *p;
  int i,x,y;
  
  cout<<"writing "<<fname<<" type is "<<_type<<endl;
  if (_type != RGB) toRGB();
  if ( (file = TIFFOpen(fname, "w")) == NULL ) {
    printf("Can't create %s!\n", fname);
    return;
    //exit(1);
  }
  
  TIFFSetField(file, TIFFTAG_IMAGEWIDTH, (uint32) _width);
  TIFFSetField(file, TIFFTAG_IMAGELENGTH, (uint32) _height);
  TIFFSetField(file, TIFFTAG_BITSPERSAMPLE, (uint32) _bitspersample);
	//  TIFFSetField(file, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
  TIFFSetField(file, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
  TIFFSetField(file, TIFFTAG_SAMPLESPERPIXEL, (uint32) _samplesperpixel);
  TIFFSetField(file, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  TIFFSetField(file, TIFFTAG_ROWSPERSTRIP, 1);
  p = new float[_height*_width*4]; 
  i=0;
  for (y=0; y<_height; y++)
    for (x=0; x<_width; x++) {
      p[i++]=(data[4*x+4*y*_width]);
      p[i++]=(data[4*x+4*y*_width+1]);
      p[i++]=(data[4*x+4*y*_width+2]);
      p[i++]=(data[4*x+4*y*_width+3]);
    }
  cout << "we got p setup\n";
  fflush(stdout);

  for (i = _height - 1; i >= 0; i--) {
    if (TIFFWriteScanline(file, p, i, 0) < 0) {
      TIFFClose(file);
      cout<<"error in write"<<endl;
    }
    p += _width * 4;
  }

  
  TIFFClose(file);
    
  delete p;
}



void GdTiff::write(char *fname) {
#ifdef OUTPUT_ALPHA
  int samples_per_pixel = 4;
#else
  int samples_per_pixel = 3;
#endif
  TIFF *file;
  unsigned char *p, *buf;
  int i=0,x,y;
  int btype = _type;
  //cout<<"writing "<<fname<<" type is "<<_type<<endl;
  if (_type != RGB) toRGB();
  if ( (file = TIFFOpen(fname, "w")) == NULL ) {
    printf("Can't create %s!\n", fname);
    return;
    //exit(1);
  }
  
  TIFFSetField(file, TIFFTAG_IMAGEWIDTH, (uint32) _width);
  TIFFSetField(file, TIFFTAG_IMAGELENGTH, (uint32) _height);
  TIFFSetField(file, TIFFTAG_BITSPERSAMPLE, 8);
  //TIFFSetField(file, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
  TIFFSetField(file, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
  TIFFSetField(file, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  TIFFSetField(file, TIFFTAG_ROWSPERSTRIP, 1);
  TIFFSetField(file, TIFFTAG_SAMPLESPERPIXEL, samples_per_pixel);
  buf = p = new unsigned char[_height*_width*samples_per_pixel]; 
  for (y=0; y<_height; y++)
    for (x=0; x<_width; x++) {
      //data[4*x+4*y*_width] = sqrt(data[4*x+4*y*_width]);
      //data[4*x+4*y*_width+1] = sqrt(data[4*x+4*y*_width+1]);
      //data[4*x+4*y*_width+2] = sqrt(data[4*x+4*y*_width+2]);
      if (data[4*x+4*y*_width] > 1.0) data[4*x+4*y*_width] = 1.0;
      if (data[4*x+4*y*_width] < 0.0) data[4*x+4*y*_width] = 0.0;
      if (data[4*x+4*y*_width+1] > 1.0) data[4*x+4*y*_width+1] = 1.0;
      if (data[4*x+4*y*_width+1] < 0.0) data[4*x+4*y*_width+1] = 0.0;
      if (data[4*x+4*y*_width+2] > 1.0) data[4*x+4*y*_width+2] = 1.0;
      if (data[4*x+4*y*_width+2] < 0.0) data[4*x+4*y*_width+2] = 0.0;
      if (data[4*x+4*y*_width+3] > 1.0) data[4*x+4*y*_width+3] = 1.0;
      if (data[4*x+4*y*_width+3] < 0.0) data[4*x+4*y*_width+3] = 0.0;
      p[i++]=(unsigned char) (data[4*x+4*y*_width]*255.0);
      p[i++]=(unsigned char) (data[4*x+4*y*_width+1]*255.0);
      p[i++]=(unsigned char) (data[4*x+4*y*_width+2]*255.0);
#ifdef OUTPUT_ALPHA
      p[i++]=(unsigned char) (data[4*x+4*y*_width+3]*255.0);
#endif
    }
  for (i = _height - 1; i >= 0; i--) {
    if (TIFFWriteScanline(file, p, i, 0) < 0) {
      TIFFClose(file);
      cout<<"error in write"<<endl;
    }
    p += _width * sizeof(unsigned char) * samples_per_pixel;
  }
  TIFFClose(file);
  switch(btype) {
  case RGB:
    break;
  case HSV:
    toHSV();
    break;
  case LAB:
    toLAB();
    break;
  }
    
  delete buf;
}

void GdTiff::toHSV() {
  double r,g,b,h,s,v,d,u,delta;
  int x,y;
  if (_type == HSV)
    return;
  if (_type == LAB)
    toRGB();
  for (y=0; y<_height; y++)
    for (x=0; x<_width; x++) {
      r = (*this)(x,y,0);
      g = (*this)(x,y,1);
      b = (*this)(x,y,2);
      if (r>g && r>b) u=r;
      else if (g>b) u=g;
      else u=b;
      if (r<g && r<b) d=r;
      else if (g<b) d=g;
      else d=b;
      v=u;
      s=(u != 0.0)?((u-d)/u):0.0;
      if (s<0.00000000001) h=0.0;
      else {
	delta=u-d;
	if (r<u+0.000000001 && r>u-0.0000000001)
	  h=(g-b)/delta;
	else if (g<u+0.000000001 && g>u-0.0000000001)
	  h=2.0+(b-r)/delta;
	else if (b<u+0.000000001 && b>u-0.0000000001)
	  h = 4.0 +(r-g)/delta;
	else {
	  cerr<<"error in gdTiff::toHSV()"<<endl;
	  exit(-1);
	}
	h*=60;
	if (h<0.0) h+=360;
      }
      (*this)(x,y,0)=h;
      (*this)(x,y,1)=s;
      (*this)(x,y,2)=v;
    }
  _type = HSV;
}

void GdTiff::toRGB() {
  int x,y,i;
  if (_type==UNDEF) {
    cerr<<"toRGB() called on image of undefined type, assuming RGB"<<endl;
    return;
  }
  if (_type==RGB)
    return;
  if (_type==HSV) {
    double r,g,b,h,s,v,f,p,q,t;
    for (y=0; y<_height; y++)
      for (x=0; x<_width; x++) {
	h=(*this)(x,y,0);
	s=(*this)(x,y,1);
	v=(*this)(x,y,2);
	if (s < 0.00000000001) {
	  r = v;
	  g = v;
	  b = v;
	} else {
	  if (h >= 360.0)
	    h -=360.0;
	  h/=60.0;
	  i = (int)floor(h);
	  f = h-i;
	  p = v*(1.0-s);
	  q = v*(1.0 - (s*f));
	  t = v*(1.0-(s*(1.0-f)));
	  switch(i) {
	  case 0: r=v; g=t; b=p; break;
	  case 1: r=q; g=v; b=p; break;
	  case 2: r=p; g=v; b=t; break;
	  case 3: r=p; g=q; b=v; break;
	  case 4: r=t; g=p; b=v; break;
	  case 5: r=v; g=p; b=q; break;
	  }
	}
	(*this)(x,y,0)=r;
	(*this)(x,y,1)=g;
	(*this)(x,y,2)=b;
      }
    } else if (_type==LAB) {
      double l,a,b,R,G,B,X,Y,Z,XT,YT,ZT,fX,fY,fZ;
      for (y=0; y<_height; y++)
	for (x=0; x<_width; x++) {
	  l = (*this)(x,y,0);
	  a = (*this)(x,y,1);
	  b = (*this)(x,y,2);
	  fY = pow(((l+16.0)/116.0),3.0);
	  if (fY > 0.008856) YT=1.0; else YT=0.0;
	  fY = (1.0-YT)*(l/903.3)+YT*fY;
	  Y = fY;
	  fY = YT*(pow(fY,1.0/3.0))+(1.0-YT)*(7.787*fY+16.0/116.0);
	  fX = a/500.0 + fY;
	  if (fX > 0.206893) XT=1.0; else XT=0.0;
	  X = XT*pow(fX,3.0)+(1.0-XT)*((fX-16.0/116.0)/7.787);
	  fZ = fY - b/200.0;
	  if (fZ > 0.206893) ZT=1.0; else ZT=0.0;
	  Z = ZT*pow(fZ,3.0)+(1.0-ZT)*((fZ-16.0/116.0)/7.787);
	  X *= 0.950456;
	  Z *= 1.088754;
	  R = 3.240479*X+(-1.537150)*Y+(-0.498535)*Z;
	  G = (-0.969256)*X+1.875992*Y+0.041556*Z;
	  B = 0.055648*X+(-0.204043)*Y+1.057311*Z;
	  if (R < 0.0) R = 0.0;
	  if (R > 1.0) R = 1.0;
	  if (G < 0.0) G = 0.0;
	  if (G > 1.0) G = 1.0;
	  if (B < 0.0) B = 0.0;
	  if (B > 1.0) B = 1.0;
	  (*this)(x,y,0)=R;
	  (*this)(x,y,1)=G;
	  (*this)(x,y,2)=B;
	}
    }
    _type = RGB;
}

void GdTiff::toLAB() {
  double l,a,b,R,G,B,X,Y,Z, XT, YT, ZT, fX, fY, fZ;
  int x,y;
  double maxL=-1000.0, maxA=-1000.0, maxB=-1000.0;
  double minL=1000.0, minA=1000.0, minB=1000.0;
  //cout<<"converting to LAB"<<endl;
  if (_type==LAB) return;
  if (_type!=RGB) toRGB();
  for (y=0; y<_height; y++)
    for (x=0; x<_width; x++) {
      R = (*this)(x,y,0);
      G = (*this)(x,y,1);
      B = (*this)(x,y,2);
      X = 0.412453*R+0.357580*G+0.180423*B;
      Y = 0.212671*R+0.715160*G+0.072169*B;
      Z = 0.019334*R+0.119193*G+0.950227*B;
      X /= 0.950456;
      Z /= 1.088754;
      if (X>0.008856) XT = 1.0; else XT = 0.0;
      if (Y>0.008856) YT = 1.0; else YT = 0.0;
      if (Z>0.008856) ZT = 1.0; else ZT = 0.0;
      fX = XT*pow(X,1.0/3.0) + (1-XT)*(7.787*X+16.0/116.0);
      fY = YT*pow(Y,1.0/3.0) + (1-YT)*(7.787*Y+16.0/116.0);
      fZ = ZT*pow(Z,1.0/3.0) + (1-ZT)*(7.787*Z+16.0/116.0);
      l = YT*(116.0*pow(Y,1.0/3.0)-16.0) + (1.0-YT)*(903.3*Y);
      a = 500.0*(fX-fY);
      b = 200.0*(fY-fZ);
      if (l > maxL) maxL=l;
      if (a > maxA) maxA=a;
      if (b > maxB) maxB=b;
      if (l < minL) minL=l;
      if (a < minA) minA=a;
      if (b < minB) minB=b;
      (*this)(x,y,0)=l;
      (*this)(x,y,1)=a;
      (*this)(x,y,2)=b;
    }
  //cout<<"max ("<<maxL<<","<<maxA<<","<<maxB<<")"<<endl;
  //cout<<"min ("<<minL<<","<<minA<<","<<minB<<")"<<endl;
  _type = LAB;
}


/*
int main (int argc, char **argv) {
  int x,y,c;
  cout<<"about to read tif"<<endl;
  GdTiff tiff(argv[1]);
  cout<<"about to create ppm"<<endl;
  GdTiff ppm(tiff.width(), tiff.height());
  cout<<"about to copy"<<endl;
  for (y=0; y<tiff.height(); y++)
    for (x=0; x<tiff.width(); x++)
      for (c=0; c<4; c++)
	if (tiff(x,y,c) > 1.0)
	  ppm(x,y,c) = 1.0;
	else if (tiff(x,y,c) < 0.0)
	  ppm(x,y,c) = 0.0;
	else
	  ppm(x,y,c) = tiff(x,y,c);
  cout<<"about to write ppm"<<endl;
  ppm.write(argv[2]);
  cout<<"exiting..."<<endl;
}
*/

float** readChannel(char* fname, int &width, int &height) {
  int i,x,y;
  float **z;
  ifstream din(fname, ios::in|ios::binary);
  din.read((char*)&i, sizeof(int));
  char ch;
  width = height = 0;
  din.read(&ch, 1);
  width+=(int)ch;
  din.read(&ch, 1);
  width+=256*(int)ch;
  din.read(&ch, 1);
  height+=(int)ch;
  din.read(&ch, 1);
  height+=256*(int)ch;
  float t;
  for (i=0; i<32; i++)
    din.read((char*)&t,sizeof(float));
  z=new float*[height];
  for (i=0; i<height; i++)
    z[i] = new float[width];
  for (y=0; y<height; y++) 
    for (x=0; x<width; x++) 
      din.read((char*)&(z[y][x]), sizeof(float));
  return z;
}
